/***************************************************************************/
/*
  Project: Power Estimation for FPGA
  Institue: The University of British Columbia
  Developer: Kara Poon
  Supervisor: Dr. Steve Wilton

  This source file serves as an add-on feature to the original VPR program. 
  It compute the dynamic, short-circuit, and leakage power dissipation based
  on the activities estimated by the activity generator.

  Three output files will be produced:
  ROUTING_power.echo : contains routing power dissipation for each net
  LBpower.echo       : contains logic block power dissipation for each LUT
  CLKpower.echo      : contains the clock power dissiaption 
  power.echo         : contains the summary of power dissipation for this
                       circuit

 */

/* external function */
float get_power_per_clock_cycle
(
 t_subblock_data *subblock_data_ptr,
 int num_switch, 
 float T_crit, 
 int num_segment,
 t_segment_inf *segment_inf, 
 int total_segment_used,
 enum e_switch_block_type switch_block_type
);

/* 
  Function: Calculate the power of the circuit

  variables: 
	*subblock_data_ptr: a pointer to a subblock data record
	num_switch: number of different type of switches
	T_crit:     Timing for the critical path
	num_segment: number of different type of segments	
	*segment_inf: Pointer to a segment record
	switch_block_type:  switch block type (subset, Universal, Wilton, Wilton3) 
*/


void find_cap
(
 tx_type transistor_type,
 float length,
 float width,
 float* C_drain,
 float* C_source,
 float* C_gate
);

/*
  Function : Find the capacitance of a transistor
  output: Drain, source, and gate capacitance
  variables:
  	length: channel length of the transistor
  	width: channel width of the transistor   
  	C_drain: Drain capacitance
  	C_source: Source capacitance
  	Cgate: Gate capaciatance
*/


void calculate_INV_cap
(
 float PMOS_length,
 float PMOS_width,
 float NMOS_length,
 float NMOS_width,
 float *Cin,
 float *Cout
);

/*
  Function : Calculate the input and output capacitance of an inverter
  output: input and output capacitance of the inverter
  variables: 
  	PMOS_length: transistor length of the PMOS
  	PMOS_width:  transistor width of the PMOS
  	NMOS_length: transistor length of the NMOS
  	NMOS_width:  transistor length of the NMOS
  	*Cin: pointer to input capacitance of the inverter
  	*Cout: pointer to output capacitance of the inverter
*/
