/***********************************************************************
	Read the constraint file for VEB
***********************************************************************/
#include <string.h>
#include <stdio.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "read_constraint.h"
#include <assert.h>



/******************************************************************************
	Read the link file to create index for all modules
	Read the area constraint of each module
 ******************************************************************************/
void create_module_index(char * veb_link_file_name)
{
	
	FILE *fp_veb_link;
	FILE *fp_veb;
	char veb_link_buf[BUFSIZE];
	char veb_buf[BUFSIZE];
	char *module_name;
	char *veb_link_ptr;
	char *temp_ptr;
	int i;
	

	number_module = 0;

	//open the link file to get the net file for veb
	fp_veb_link = my_fopen(veb_link_file_name, "r", 0);
	// get a line in the link file
	veb_link_ptr = my_fgets(veb_link_buf, BUFSIZE, fp_veb_link);

	//first pass, read the link file to find the top module
	while(veb_link_ptr != NULL)
	{
		temp_ptr = my_strtok(veb_link_buf, TOKENS, fp_veb_link, veb_link_buf);
		
		if ((temp_ptr != NULL)  && (strcmp(temp_ptr,".top") == 0) )
		{
			//get the top module
			//create new index list
			constraint = malloc (sizeof(struct user_constraint));
			
			//get the top module name 
			temp_ptr = my_strtok(NULL, TOKENS, fp_veb_link, veb_link_buf);
			constraint = malloc(sizeof(struct user_constraint));
			constraint[0].name = malloc((strlen(temp_ptr)+1)*sizeof(char));
			strcpy(constraint[0].name, temp_ptr);			

			//get the ucf file name
			temp_ptr = my_strtok(NULL, TOKENS, fp_veb_link, veb_link_buf);
			temp_ptr = my_strtok(NULL, TOKENS, fp_veb_link, veb_link_buf);
			if (temp_ptr == NULL)
			{
				//no constraint file
				constraint[0].ucf_name = NULL;
			}
			else
			{
				constraint[0].ucf_name = malloc((strlen(temp_ptr)+1)*sizeof(char));
				strcpy(constraint[0].ucf_name, temp_ptr);			
			}

			number_module ++;

			//initial the value
			constraint[0].width = -1;
			constraint[0].height = -1;
			constraint[0].num_block_list = 0;
			constraint[0].lx = -1;
			constraint[0].ly = -1;
			constraint[0].veb_index = 0;
			constraint[0].constraint_lower_corner=0;
			break;
		}
		else
		{
			veb_link_ptr = my_fgets(veb_link_buf, BUFSIZE, fp_veb_link);
		}
	}

	if (number_module ==0)
	{
		printf("Error in reading constraint ! \n");
		printf("Cannot find top module name\n");
		exit(1);
	}

	//second pass, read the link file to find the veb module

	rewind(fp_veb_link);

	veb_link_ptr = my_fgets(veb_link_buf, BUFSIZE, fp_veb_link);
	while(veb_link_ptr != NULL)
	{
		temp_ptr = my_strtok(veb_link_buf, TOKENS, fp_veb_link, veb_link_buf);
		if ((temp_ptr != NULL) && (strcmp(temp_ptr,".top") != 0))
		{
			//new VEB module
			number_module ++;	//increase the number of number module
			constraint = realloc (constraint, number_module  * sizeof(struct user_constraint));
			//copy the module name
			constraint[number_module-1].name = malloc((strlen(temp_ptr)+1)*sizeof(char));
			strcpy(constraint[number_module-1].name, temp_ptr);			

			//get the ucf file name
			temp_ptr = my_strtok(NULL, TOKENS, fp_veb_link, veb_link_buf);
			temp_ptr = my_strtok(NULL, TOKENS, fp_veb_link, veb_link_buf);

			if (temp_ptr == NULL)
			{
				constraint[number_module-1].ucf_name = NULL;
			}
			else
			{
				constraint[number_module-1].ucf_name = malloc((strlen(temp_ptr)+1)*sizeof(char));
				strcpy(constraint[number_module-1].ucf_name, temp_ptr);	
			}

			//initial the value
			constraint[number_module-1].width = -1;
			constraint[number_module-1].height = -1;
			constraint[number_module-1].num_block_list = 0;
			constraint[number_module-1].lx = -1;
			constraint[number_module-1].ly = -1;
			constraint[number_module-1].veb_index = number_module-1;
			constraint[number_module-1].constraint_lower_corner=0;

		}
		veb_link_ptr = my_fgets(veb_link_buf, BUFSIZE, fp_veb_link);
	}

	fclose(fp_veb_link);

	return;
}

/******************************************************************************
	Get the veb index by giving the name
 ******************************************************************************/
int get_module_index(char * veb_name)
{
	int count=0;

	for(count=0;count<number_module;count++)
	{	
		if (strcmp(veb_name, constraint[count].name) == 0)
		{			
			return count;
		}
	}

	return(-1);
}


/********************************************************************************
	Set user constraint to each block
 ********************************************************************************/
void add_block_constraint()
{
	int i,j;
	FILE *fp_constraint;
	char *constraint_pointer;
	char constraint_buf[BUFSIZE];
	char *temp_ptr;
	char *clb_name;
	int k;
	int type;

	int valid, iclass, ipin;

	char *position[4] = {"top", "bottom", "left", "right"};	

	printf("Adding constraint \n");
	
	for (i=0;i<number_module;i++)
	{
		if (constraint[i].ucf_name != NULL)
		{
			printf("Reading ucf file name : %s \n", constraint[i].ucf_name);

			//there is ucf file for that module
			fp_constraint = my_fopen (constraint[i].ucf_name, "r", 0);
			constraint_pointer = my_fgets(constraint_buf, BUFSIZE, fp_constraint);															

			constraint[i].fixed = 0;
			constraint[i].type = 0;

			while (constraint_pointer != NULL)
			{
				temp_ptr = my_strtok(constraint_pointer, TOKENS, fp_constraint, constraint_buf);
				

				if (temp_ptr != NULL)
				{
					//Check the constraint
					if (strcmp(temp_ptr, "WIDTH") == 0)
					{
						//Width constraint (in term of number of clb
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

						if (temp_ptr != NULL)
						{
							constraint[i].width = atoi(temp_ptr);
							printf("Width : %d \n", constraint[i].width);
						}
						else
						{
							printf("Error in reading width constraint of file %s \n", constraint[i].ucf_name);
							exit(1);
						}
					}
					else if (strcmp(temp_ptr, "HEIGHT") == 0)
					{
						//Height constraint (in term of number of clb
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

						if (temp_ptr != NULL)
						{
							constraint[i].height = atoi(temp_ptr);
							printf("Height : %d \n", constraint[i].height);
						}
						else
						{
							printf("Error in reading height constraint of file %s \n", constraint[i].ucf_name);
							exit(1);
						}
					}
					else if(strcmp(temp_ptr,"FIXLOC") == 0)
					{
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

						if (temp_ptr != NULL)
						{
							constraint[i].fixed = atoi(temp_ptr);
						}

					}
					else if(strcmp(temp_ptr,"TYPE") == 0)
					{
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

						if (temp_ptr != NULL)
						{
							constraint[i].type = atoi(temp_ptr);
						}

					}
					else if(strcmp(temp_ptr,"PINLOC") == 0)
					{
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);
						
						
						if (temp_ptr != NULL)
						{
							constraint[i].num_pins = atoi(temp_ptr);
							
							constraint[i].pinside = malloc(constraint[i].num_pins * sizeof (int));
							constraint[i].pinpos_x = malloc(constraint[i].num_pins * sizeof (int));
							constraint[i].pinpos_y = malloc(constraint[i].num_pins * sizeof (int));
							constraint[i].pinpos_ulength = malloc(constraint[i].num_pins * sizeof (int));							
							constraint[i].is_global_pin = malloc(constraint[i].num_pins * sizeof (int));
							constraint[i].pin_cap = malloc(constraint[i].num_pins * sizeof (double));
							constraint[i].pin_del = malloc(constraint[i].num_pins * sizeof (double));
							constraint[i].is_outpin = malloc(constraint[i].num_pins * sizeof (int));							
							constraint[i].class_num = malloc(constraint[i].num_pins * sizeof (int));	
							constraint[i].class_inf_num_pins =  malloc(constraint[i].num_pins * sizeof (int));	
							constraint[i].pinloc = malloc(4 * sizeof (int*));	
							constraint[i].pinloc[0] = malloc(constraint[i].num_pins * sizeof (int));	
							constraint[i].pinloc[1] = malloc(constraint[i].num_pins * sizeof (int));	
							constraint[i].pinloc[2] = malloc(constraint[i].num_pins * sizeof (int));	
							constraint[i].pinloc[3] = malloc(constraint[i].num_pins * sizeof (int));	


							for (k = 0; k < constraint[i].num_pins; k++)
							{
								constraint[i].pinside[k] = -1;
								constraint[i].pinpos_x[k] = -1;
								constraint[i].pinpos_y[k] = -1;
								constraint[i].pinpos_ulength[k] = -1;
								constraint[i].is_global_pin[k] = FALSE;
								constraint[i].class_inf_num_pins[k] = 1;
								constraint[i].pinloc[0][k] = 0;
								constraint[i].pinloc[1][k] = 0;
								constraint[i].pinloc[2][k] = 0;
								constraint[i].pinloc[3][k] = 0;
							}
							printf("num_pins : %d \n", constraint[i].num_pins);							
						}
						else
						{
							printf("Error in reading PINLOC constraint of file %s \n", constraint[i].ucf_name);
							exit(1);
						}

						//get TOP number of pin
						constraint_pointer = my_fgets(constraint_buf, BUFSIZE, fp_constraint);
						temp_ptr = my_strtok(constraint_buf, TOKENS, fp_constraint, constraint_buf);	//get input or output
						if(strcmp(temp_ptr,"TOP") == 0)
						{
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

							if (temp_ptr != NULL)
							{
								constraint[i].num_pin_in_side[TOP] = atoi(temp_ptr);												
							}
							else
							{
								printf("Error in reading PINLOC constraint of file %s \n", constraint[i].ucf_name);
								exit(1);
							}
						}

						//get bottom number of pin
						constraint_pointer = my_fgets(constraint_buf, BUFSIZE, fp_constraint);
						temp_ptr = my_strtok(constraint_buf, TOKENS, fp_constraint, constraint_buf);	//get input or output
						if(strcmp(temp_ptr,"BOTTOM") == 0)
						{
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

							if (temp_ptr != NULL)
							{
								constraint[i].num_pin_in_side[BOTTOM] = atoi(temp_ptr);												
							}
							else
							{
								printf("Error in reading PINLOC constraint of file %s \n", constraint[i].ucf_name);
								exit(1);
							}
						}

						//get left number of pin
						constraint_pointer = my_fgets(constraint_buf, BUFSIZE, fp_constraint);
						temp_ptr = my_strtok(constraint_buf, TOKENS, fp_constraint, constraint_buf);	//get input or output
						if(strcmp(temp_ptr,"LEFT") == 0)
						{
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

							if (temp_ptr != NULL)
							{
								constraint[i].num_pin_in_side[LEFT] = atoi(temp_ptr);												
							}
							else
							{
								printf("Error in reading PINLOC constraint of file %s \n", constraint[i].ucf_name);
								exit(1);
							}
						}

						//get right number of pin
						constraint_pointer = my_fgets(constraint_buf, BUFSIZE, fp_constraint);
						temp_ptr = my_strtok(constraint_buf, TOKENS, fp_constraint, constraint_buf);	//get input or output
						if(strcmp(temp_ptr,"RIGHT") == 0)
						{
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

							if (temp_ptr != NULL)
							{
								constraint[i].num_pin_in_side[RIGHT] = atoi(temp_ptr);												
							}
							else
							{
								printf("Error in reading PINLOC constraint of file %s \n", constraint[i].ucf_name);
								exit(1);
							}
						}


						//printf("getting pin information \n");
						
						//get the pin information 
						for (k= 0; k< constraint[i].num_pins; k++)
						{
							constraint_pointer = my_fgets(constraint_buf, BUFSIZE, fp_constraint);

							//printf("%s", constraint_pointer);
							//printf("%s", constraint_buf);
							temp_ptr = my_strtok(constraint_buf, TOKENS, fp_constraint, constraint_buf);	//get input or output
							

							if (strcmp(temp_ptr, "emb_inpin") == 0)
							{
								//receiver
								type = RECEIVER;
								constraint[i].is_outpin[k] = FALSE;
							}
							else
							{
								type = DRIVER;
								constraint[i].is_outpin[k] =  TRUE;
							}
							
							constraint[i].class_num[k] = k;
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);
							
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);
							
							
							//skip class, every pin is a class
							//get cap 
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

							if (strcmp (temp_ptr, "cap:") != 0) 
							{
								printf("Error in get_pin_cap on line %d of architecture file.\n",linenum);
								printf("Expected cap: keyword.\n");
								exit (1);
							}

							//get cap value
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);
							
							if (temp_ptr == NULL) {
								printf("Error in get_pin_cap on line %d of architecture file.\n",linenum);
								printf("Expected pin_cap value.\n");
								exit (1);
							}
							

							constraint[i].pin_cap[k] = (double) atof (temp_ptr);

							//get global for RECEIVER
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);
							//printf("%s \n", temp_ptr);
							
							if (type == RECEIVER && strcmp (temp_ptr, "global") == 0) 
							{
								constraint[i].is_global_pin[k] = TRUE;
								
								
								temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

								//printf("%s \n", temp_ptr);
								if (temp_ptr == NULL) 
								{
									printf("Error:  emb pin statement specifies no locations, line %d.\n",
										linenum);
									exit(1);
								}
							}
														
							//check the position
							if (strcmp(temp_ptr, "top") ==0)
							{
								constraint[i].pinloc[0][k] ++;
							} 					
							else if (strcmp(temp_ptr, "bottom") ==0)							
							{
								constraint[i].pinloc[1][k] ++;
							}
							else if (strcmp(temp_ptr, "left") ==0)
							{
								constraint[i].pinloc[2][k] ++;
							}
							else if (strcmp(temp_ptr, "right") ==0)
							{
								constraint[i].pinloc[3][k] ++;
							}
							else
							{
								printf("Error:  bad pin location in ucf.\n");
								exit(1);
							}

							//get the delay deviated from internal delay, an optional parameter
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

							if (temp_ptr == NULL)
							{
								//if not specified , set to 0
								constraint[i].pin_del[k] = 0.0;

							}
							else if (strcmp (temp_ptr, "del:") == 0) 
							{
								//get del value
								temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);
								constraint[i].pin_del[k] = (double) atof (temp_ptr);
							}
																										
						}

						//get the timing of emb
						constraint_pointer = my_fgets(constraint_buf, BUFSIZE, fp_constraint);							
						temp_ptr = my_strtok(constraint_buf, TOKENS, fp_constraint, constraint_buf);	//get input or output
						
						if (strcmp(temp_ptr, "T_emb") == 0) 
						{
							//get the T_internal 
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get internal delay
							if (strcmp(temp_ptr, "T_internal:") == 0)
							{
								temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get internal delay
								constraint[i].T_emb.T_comb = (double) atof (temp_ptr);
							}

							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get emb ipin delay

							if (strcmp(temp_ptr, "T_in:") == 0)
							{
								temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get internal delay
								constraint[i].T_emb.T_seq_in = (double) atof (temp_ptr);
							}

							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get emb out pin delay
							if (strcmp(temp_ptr, "T_out:") == 0)
							{
								temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get internal delay
								constraint[i].T_emb.T_seq_out = (double) atof (temp_ptr);
							}
						}										
					}					
															
					else if(strcmp(temp_ptr, "RLOC") == 0)
					{						
						//The location constraint of normal CLB
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get the name of the clb to be constrainted
						if (temp_ptr == NULL)
						{
							printf("Error in reading RLOC constraint in file : %s \n", constraint[i].ucf_name);
							printf("Cannot find the name of component after RLOC \n");
							exit(1);
						}


						if (constraint[i].veb_index == 0)
						{
							//it is the constraint of the top module's clb 
							clb_name = malloc( (strlen(temp_ptr) + 1) * sizeof(char));
							strcpy(clb_name, temp_ptr);
							
						}
						else
						{
							//it is the constraint of the veb clb
							clb_name = malloc( (strlen(constraint[i].name) + strlen(temp_ptr) + 2) * sizeof(char));
							strcpy(clb_name, constraint[i].name);
							strcat(clb_name, "_\0");
							strcat(clb_name, temp_ptr);
						}

						printf("finding block match RLOC name \n");

						for (j=0;j<num_blocks;j++)
						{
							//find the block which has the same name as constraint
							if (block[j].module_index == constraint[i].veb_index )
							{
								if (strcmp(block[j].name, clb_name) ==0)
								{
									if (constraint[i].num_block_list ==0)
									{
										constraint[i].block_list = malloc(sizeof(int));
									}
									else
									{
										constraint[i].block_list = realloc(constraint[i].block_list, (constraint[i].num_block_list + 1) * sizeof(int));
									}
									constraint[i].block_list[constraint[i].num_block_list] = j;
									constraint[i].num_block_list ++;


									//get the x position of the clb
									temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);
									if (temp_ptr != NULL)
									{
										block[j].rx = atoi(temp_ptr);
									}
									else
									{
										printf("Error in reading RLOC constraint in file : %s \n", constraint[i].ucf_name);
										printf("Cannot find the x position value of component after RLOC \n");
									}

									//get the y position of the clb
									temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);
									if (temp_ptr != NULL)
									{
										block[j].ry = atoi(temp_ptr);
									}
									else
									{
										printf("Error in reading RLOC constraint in file : %s \n", constraint[i].ucf_name);
										printf("Cannot find the y position value of component after RLOC \n");
									}
									if (constraint[i].veb_index == 0)
									{
										printf("block[%d] rx =%d, ry=%d\n", j, block[j].rx, block[j].ry);										
									}
									j = num_blocks;
								}

							}
						}
						
						free(clb_name);
						
					}
					else if(strcmp(temp_ptr,"DSPLOC") ==0)
					{
						//get which column the dsp is placed
						
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get the lower x position of the veb 
						while (temp_ptr != NULL)
						{
							if (dsp_inf->num_pos ==0)
							{
								dsp_inf->dsp_pos = malloc(sizeof(int));
							}
							else
							{
								dsp_inf->dsp_pos = (int *)realloc(dsp_inf->dsp_pos, (dsp_inf->num_pos+1)*sizeof(int));
							}
							(dsp_inf->dsp_pos)[dsp_inf->num_pos] = atoi(temp_ptr);
							dsp_inf->num_pos ++;

							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get the lower x position of the veb 
						}

						
					}
					else if( strcmp(temp_ptr,"MEMLOC") ==0)
					{
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get the lower x position of the veb 
						while (temp_ptr != NULL)
						{
							if (memory_inf->num_pos ==0)
							{
								memory_inf->memory_pos = malloc(sizeof(int));
							}
							else
							{
								memory_inf->memory_pos = (int *)realloc(memory_inf->memory_pos, (memory_inf->num_pos+1)*sizeof(int));
							}
							(memory_inf->memory_pos)[memory_inf->num_pos] = atoi(temp_ptr);
							memory_inf->num_pos ++;

							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get the lower x position of the veb 
						}
					}

					else if(strcmp(temp_ptr,"VEBLOC") ==0 || strcmp(temp_ptr,"EMBLOC") ==0)
					{
						//The location constraint of the VEB
						if (constraint[i].veb_index == 0)
						{
							//it is the top module. It can have constraint of lower x,y position of VEB in the whole FPGA
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get the name of the veb to be constrainted

							for (j= 0;j<number_module; j++)
							{
								if (strcmp(temp_ptr, constraint[j].name) == 0)
								{
									temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get the lower x position of the veb 
									if (temp_ptr == NULL)
									{
										printf("Error in reading VEBLOC constraint in file : %s \n", constraint[i].ucf_name);
										printf("Cannot find the x position value of VEB after VEBLOC \n");
									}

									constraint[j].lx = atoi(temp_ptr);

									
									temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);	//get the lower y position of the veb 
									if (temp_ptr == NULL)
									{
										printf("Error in reading VEBLOC constraint in file : %s \n", constraint[i].ucf_name);
										printf("Cannot find the y position value of VEB after VEBLOC \n");
									}

									constraint[j].ly = atoi(temp_ptr);

									constraint[j].constraint_lower_corner=1;

									j = number_module;
								}
							}																						
						}
						else
						{
							//it is VEB, can't have constraint on the lower x,y position of VEB in the whole FPGA
							printf("Error in reading constraint in file : %s \n", constraint[i].ucf_name);
							printf("Cannot have VEBRLOC in veb ucf file ! \n");
						}
					}
					else if (strcmp(temp_ptr, "EMB_SWITCH") == 0)
					{
						//set emb_allow_switch
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

						if (temp_ptr != NULL)
						{
							emb_allow_switch = atoi(temp_ptr);
							printf("EMB_SWITCH : %d \n", emb_allow_switch);
						}
						else
						{
							printf("Error in reading EMB_SWITCH constraint of file %s \n", constraint[i].ucf_name);
							exit(1);
						}
					}					
					else if (strcmp(temp_ptr, "WIDESEG_LOC") == 0)
					{
						//set emb_allow_switch
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

						if (temp_ptr != NULL)
						{
							//read the x position of wide segment
							wide_seg.x = atoi(temp_ptr);
							temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);
							if (temp_ptr != NULL)
							{
								wide_seg.y = atoi(temp_ptr);
								printf("WIDE SEGMENT LOCATION : %d, %d \n", wide_seg.x, wide_seg.y);
							}
							else
							{
								printf("Error in reading WIDESEG_LOC constraint of file %s \n", constraint[i].ucf_name);
								exit(1);
							}							
						}
						else
						{																					
							printf("Error in reading WIDESEG_LOC constraint of file %s \n", constraint[i].ucf_name);
							exit(1);
						}
					}					
					else if (strcmp(temp_ptr, "WIDESEG_HEIGHT") == 0)
					{
						//set the height of wide segment
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

						if (temp_ptr != NULL)
						{
							wide_seg.height = atoi(temp_ptr);
							printf("WIDESEG_HEIGHT : %d \n", wide_seg.height);
						}
						else
						{
							printf("Error in reading WIDESEG_HEIGHT constraint of file %s \n", constraint[i].ucf_name);
							exit(1);
						}
					}					
					else if (strcmp(temp_ptr, "WIDESEG_WIDTH") == 0)
					{
						//set the width of wide segment
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

						if (temp_ptr != NULL)
						{
							wide_seg.width = atoi(temp_ptr);
							printf("WIDESEG_WIDTH : %d \n", wide_seg.width);
						}
						else
						{
							printf("Error in reading WIDESEG_WIDTH constraint of file %s \n", constraint[i].ucf_name);
							exit(1);
						}
					}					
					else if (strcmp(temp_ptr, "WIDESEG_RATIO") == 0)
					{
						//set the height of wide segment
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

						if (temp_ptr != NULL)
						{							
							wide_seg.ratio = (double) atof (temp_ptr);
							printf("WIDESEG_RATIO : %g \n", wide_seg.ratio);
						}
						else
						{
							printf("Error in reading WIDESEG_RATIO constraint of file %s \n", constraint[i].ucf_name);
							exit(1);
						}
					}					
					else if (strcmp(temp_ptr, "IS_WIDESEG") == 0)
					{
						//set the height of wide segment
						temp_ptr = my_strtok(NULL, TOKENS, fp_constraint, constraint_buf);

						if (temp_ptr != NULL)
						{							
							is_wide_seg =  atoi (temp_ptr);
							printf("IS_WIDESEG : %d \n", is_wide_seg);
						}
						else
						{
							printf("Error in reading IS_WIDESEG constraint of file %s \n", constraint[i].ucf_name);
							exit(1);
						}
					}					
				}
				constraint_pointer = my_fgets(constraint_buf, BUFSIZE, fp_constraint);
			}

			fclose(fp_constraint);
		}
	}
}

void set_clb_to_emb()
{
	int width , height, lx , ly;
	int i,j,k;
	int iblock; 
	for (i= 1; i<number_module; i++)
	{
		width= constraint[i].width;
		height = constraint[i].height;
		lx = constraint[i].lx;
		ly = constraint[i].ly;

		
		for(iblock  =0 ; iblock < num_blocks; iblock ++)
		{
			if (block[iblock].module_index == i)
			{
				break;
			}
		}

		for (j= lx; j < lx + width; j++)
		{
			for (k = ly; k < ly + height; k++)
			{
				clb[j][k].type = EMB;
				clb[j][k].module_index = constraint[i].veb_index;
				clb[j][k].iconstraint = i;
				clb[j][k].u.block = iblock;
				//clb[j][k].occ = 1;
				//clb[j][k].empty = 0;
			}
		}

	}
}


void sort_constraint()
{
	int i, j, k;
	int temp;

	
	for (i= 0; i<number_module; i++)
	{
/*
		printf("Before sorting \n");
		printf("Component name : %s \n", constraint[i].name);
		for (j=0; j<constraint[i].num_block_list;j++)
		{
			printf("Block name %d : %s , RLOC = %d %d\n", j, block[constraint[i].block_list[j]].name, block[constraint[i].block_list[j]].rx, block[constraint[i].block_list[j]].ry);

		}
*/

		//do bubble sort to sort the location of block for lower left conner 
		
		/*
			Block :

			[1,0]	[1,1]	[1,2]

			[0,0]	[0,1]	[0,2]
			---------------------->
			
			After sorting : [0,0]	[0,1]	[0,2]	[1,0]	[1,1]	[1,2]

		*/
		for (j=0;j<constraint[i].num_block_list-1; j++)
		{
			for (k=0;k<constraint[i].num_block_list - 1- j; k++)
			{
				if (block[constraint[i].block_list[k]].rx == block[constraint[i].block_list[k+1]].rx)
				{
					//they are on the same row, then compare the column
					if (block[constraint[i].block_list[k]].ry == block[constraint[i].block_list[k+1]].ry)
					{
						//error, they shouldn't have the same location
						printf("Error in constraint\n");
						printf("Two blocks have the same location constraint\n");
						exit(1);
					}
					else if (block[constraint[i].block_list[k]].ry > block[constraint[i].block_list[k+1]].ry)
					{
						temp = constraint[i].block_list[k];
						constraint[i].block_list[k] = constraint[i].block_list[ k+ 1];
						constraint[i].block_list[ k+ 1] = temp;
					}
				}
				else if (block[constraint[i].block_list[k]].rx > block[constraint[i].block_list[k+1]].rx)
				{
					temp = constraint[i].block_list[k];
					constraint[i].block_list[k] = constraint[i].block_list[ k+ 1];
					constraint[i].block_list[ k+ 1] = temp;

				}

			}
		}

		/*
		//set emb block x, y
		for (j=0; j<constraint[i].num_block_list;j++)
		{
			block[constraint[i].block_list[j]].x = constraint[i].lx;
			block[constraint[i].block_list[j]].y = constraint[i].ly;
		}
*/

		/*
		printf("After sorting \n");
		printf("Component name : %s \n", constraint[i].name);
		
		for (j=0; j<constraint[i].num_block_list;j++)
		{
			printf("Block name %d : %s \n", j, block[constraint[i].block_list[j]].name);			
		}
		if (constraint[i].veb_index > 0)
		{
//			printf("block_list[0].x %d, block_list[0].y %d \n", block[constraint[i].block_list[0]].x, block[constraint[i].block_list[0]].y);
			printf("EMB location %d : %d \n", constraint[i].lx, constraint[i].ly);
			printf("EMB height %d , width %d \n", constraint[i].height, constraint[i].width);
		}
		*/
	}
}

void free_constraint()
{
	int i;
	for (i= 0; i<number_module; i++)
	{
		if (constraint[i].name != NULL)
		{
			free(constraint[i].name);
		}

		if (constraint[i].ucf_name != NULL)
		{
			free(constraint[i].ucf_name);
		}

		if (constraint[i].block_list != NULL)
		{
			free(constraint[i].block_list);
		}

		if (constraint[i].pinside != NULL)
		{
			free(constraint[i].pinside);
		}

		if (constraint[i].pinpos_x != NULL)
		{
			free(constraint[i].pinpos_x);
		}

		if (constraint[i].pinpos_y != NULL)
		{
			free(constraint[i].pinpos_y);
		}

		if (constraint[i].pinpos_ulength != NULL)
		{
			free(constraint[i].pinpos_ulength);
		}

		if (constraint[i].is_global_pin != NULL)
		{
			free(constraint[i].is_global_pin);
		}

		if (constraint[i].pin_cap != NULL)
		{
			free(constraint[i].pin_cap);
		}

		if (constraint[i].pin_del != NULL)
		{
			free(constraint[i].pin_del);
		}

		if (constraint[i].is_outpin != NULL)
		{
			free(constraint[i].is_outpin);
		}

		if (constraint[i].class_num != NULL)
		{
			free(constraint[i].class_num);
		}

		if (constraint[i].class_inf_num_pins != NULL)
		{
			free(constraint[i].class_inf_num_pins);
		}
		
		if (constraint[i].pinloc != NULL)
		{
			if (constraint[i].pinloc[0] != NULL)
			{
				free(constraint[i].pinloc[0]);
			}

			if (constraint[i].pinloc[1] != NULL)
			{
				free(constraint[i].pinloc[1]);
			}

			if (constraint[i].pinloc[2] != NULL)
			{
				free(constraint[i].pinloc[2]);
			}

			if (constraint[i].pinloc[3] != NULL)
			{
				free(constraint[i].pinloc[3]);
			}

			free (constraint[i].pinloc);
		}													
	}
	
	if (constraint != NULL)
	{
		free(constraint);
	}

}

