#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "vpr_utils.h"

/* This module contains subroutines that are used in several unrelated parts *
 * of VPR.  They are VPR-specific utility routines.                          */


/******************** Subroutine definitions ********************************/


boolean is_opin (int ipin) {

/* Returns TRUE if this clb pin is an output, FALSE otherwise. */

 int iclass;

 iclass = clb_pin_class[ipin];

 if (class_inf[iclass].type == DRIVER)
    return (TRUE);
 else
    return (FALSE);
}


boolean is_opin_dsp (int ipin) {

/* Returns TRUE if this clb pin is an output, FALSE otherwise. */

 int iclass;

 iclass = dsp_pin_class[ipin];

 if (class_inf_dsp[iclass].type == DRIVER)
    return (TRUE);
 else
    return (FALSE);
}

boolean is_opin_memory (int ipin) {

/* Returns TRUE if this clb pin is an output, FALSE otherwise. */

 int iclass;

 iclass = memory_pin_class[ipin];

 if (class_inf_memory[iclass].type == DRIVER)
    return (TRUE);
 else
    return (FALSE);
}


void load_one_clb_fanout_count (int subblock_lut_size, t_subblock
         *subblock_inf, int num_subblocks, int *num_uses_of_clb_ipin,
         int **num_uses_of_sblk_opin, int iblk, int subblock_num_opin,  int subblock_num_ipin) 
{

 
	/* Loads the fanout counts for one block (iblk).  */
 
	int isub, ipin, conn_pin, opin;

//	printf("load_one_clb_fanout_count: iblk = %d(%s)\n", iblk, block[iblk]);
	for (isub=0;isub<num_subblocks;isub++) 
	{
 
	  /* Is the subblock output connected to a CLB opin that actually goes *
	   * somewhere?  Necessary to check that the CLB opin connects to      *
	   * something because some logic blocks result in netlists where      *
	   * subblock outputs being automatically hooked to a CLB opin under   *
	   * all conditions.                                                   */

		for (opin = 0; opin < subblock_num_opin; opin ++)
		{
			//if the output pin of the subblock connect ot slice output and this slice output connect to a net , increase the num_uses_of_sblk_opin
			conn_pin = subblock_inf[isub].outputs[opin];		

			if (conn_pin != OPEN) 
			{

				if (conn_pin >= pins_per_clb)
				{
					if (opin != O_Y)
					{
						//internal hidden connection
						num_uses_of_sblk_opin[isub][opin]++;	
//						printf("isub: %d , opin : %d, num_uses_of_sblk_opin[%d][%d] : %d \n", isub, opin, isub, opin, num_uses_of_sblk_opin[isub][opin]);
					}
					//don't connect X/Y to DX/DY, leave them open
				}
				else
				{
					if (block[iblk].nets[conn_pin] != OPEN)
					{   /* CLB output is used */
						num_uses_of_sblk_opin[isub][opin]++;	
//						printf("isub: %d , opin : %d, num_uses_of_sblk_opin[%d][%d] : %d \n", isub, opin, isub, opin, num_uses_of_sblk_opin[isub][opin]);
					}

					//the opin X/Y
					if (opin == O_Y)
					{
						//X/Y may connect to DX/DY
						if (subblock_inf[isub].inputs[I_DY] != OPEN)	//one more edge connect to DX/DY
						{
							//num_uses_of_sblk_opin[isub][opin]++;	
							//printf("X/Y connect to DX/DY -- isub: %d , opin : %d, num_uses_of_sblk_opin[%d][%d] : %d \n", isub, opin, isub, opin, num_uses_of_sblk_opin[isub][opin]);
						}
					}
				}
				
			}
		}
 
	
		//check the input pin of each subblock
		//if the subblock input connect to net , increase num_uses_of_clb_ipin
	    for (ipin=0;ipin<subblock_num_ipin;ipin++) 
		{
    	   	conn_pin = subblock_inf[isub].inputs[ipin];
 
       		if (conn_pin != OPEN) 
			{
          		if (conn_pin < pins_per_clb) 
				{   /* Driven by CLB ipin */
             		num_uses_of_clb_ipin[conn_pin]++;
          		}
          		else 
				{
			   		/* Driven by sblk output in same clb */
	             	//num_uses_of_sblk_opin[conn_pin - pins_per_clb]++;
				 	//no futher need since the the sblk output is counted in the previous loop
        	  	}
       		}	   
    	}  /* End for each sblk ipin */

		
	    conn_pin = subblock_inf[isub].clock;  // Now do clock pin 

    	if (conn_pin != OPEN) 
		{
			if (conn_pin < pins_per_clb) 
			{  //* Driven by CLB ipin 
	        	num_uses_of_clb_ipin[conn_pin]++;
    		} 
		    else 
			{   // Driven by sblk output in same clb 
	         // num_uses_of_sblk_opin[conn_pin - pins_per_clb]++;
    	   	} 
	    }	   
		
		//printf("subblock name : %s, num_uses of clb ipin : %d, num uses of sblk opin : %d \n", subblock_inf[isub].name, num
	}  /* End for each subblock */
}
